#include <QMessageBox>
#include <QImage>

#include "multicamera.h"
#include "ui_multicamera.h"

#include "TeliCamApi.h"
#include "TeliCamUtl.h"

MultiCamera::MultiCamera(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MultiCamera)
{
    ui->setupUi(this);

    this->m_bIsOpen = false;

    CAM_API_STATUS status;

    this->pArrayGroupBox[0] = ui->groupBox1;
    this->pArrayGroupBox[1] = ui->groupBox2;
    this->pArrayGroupBox[2] = ui->groupBox3;
    this->pArrayGroupBox[3] = ui->groupBox4;

    this->pArrayDoubleSpinBoxGain[0] = ui->doubleSpinBoxGain1;
    this->pArrayDoubleSpinBoxGain[1] = ui->doubleSpinBoxGain2;
    this->pArrayDoubleSpinBoxGain[2] = ui->doubleSpinBoxGain3;
    this->pArrayDoubleSpinBoxGain[3] = ui->doubleSpinBoxGain4;

    this->pArrayDoubleSpinBoxFrameRate[0] = ui->doubleSpinBoxFrameRate1;
    this->pArrayDoubleSpinBoxFrameRate[1] = ui->doubleSpinBoxFrameRate2;
    this->pArrayDoubleSpinBoxFrameRate[2] = ui->doubleSpinBoxFrameRate3;
    this->pArrayDoubleSpinBoxFrameRate[3] = ui->doubleSpinBoxFrameRate4;

    this->pArrayLabelInfo[0] = ui->labelInfo1;
    this->pArrayLabelInfo[1] = ui->labelInfo2;
    this->pArrayLabelInfo[2] = ui->labelInfo3;
    this->pArrayLabelInfo[3] = ui->labelInfo4;

    this->pArrayLabelFps[0] = ui->labelFps1;
    this->pArrayLabelFps[1] = ui->labelFps2;
    this->pArrayLabelFps[2] = ui->labelFps3;
    this->pArrayLabelFps[3] = ui->labelFps4;

    this->pArrayLabelImage[0] = ui->labelImage1;
    this->pArrayLabelImage[1] = ui->labelImage2;
    this->pArrayLabelImage[2] = ui->labelImage3;
    this->pArrayLabelImage[3] = ui->labelImage4;

    // Clear statis flags.
    this->m_uiCamCount              = 0;
    this->m_bAcquisActive           = false;
    this->m_hImgUpdThrd             = NULL;

    // Clear variables for cameras
    for (int i = 0; i < MAX_CAMERA; ++i)
    {
        this->m_hCam[i]             = (CAM_HANDLE)NULL;
        this->m_hStrm[i]			= (CAM_STRM_HANDLE)NULL;
        this->m_pvImgBGRA[i]		= NULL;
        this->m_uiFrmCount[i]		= 0;
        this->m_uiFrmIndex[i]		= 0;
        this->m_bFrmIndexValid[i]	= false;
    }

    this->m_uiSelectedTabIndex = 0;

    // Initialize Api
    status = Sys_Initialize();
    if(status != CAM_API_STS_SUCCESS)
    {
        SetError(QString("Sys_Initialize error!(%1)").arg(status));
        return;
    }

    // Prepare table for converting YUV image to BGRA.
    status = PrepareLUT();
    if (status != CAM_API_STS_SUCCESS)
    {
        SetError(QString("PrepareLUT error!(%1)").arg(status));
        return;
    }

    for (uint32_t camNo = 0; camNo < MAX_CAMERA; camNo++)
    {
        ui->tabWidget->setTabEnabled(camNo, false);
        this->pArrayLabelInfo[camNo]->setText("");
    }

    ui->pushButtonUpdateCameraList->setEnabled(true);
    ui->pushButtonStartAcquisition->setEnabled(false);
    ui->pushButtonStopAcquisition->setEnabled(false);
}

MultiCamera::~MultiCamera()
{
    Terminate();

    delete ui;
}

void MultiCamera::SetError(QString text)
{
    QMessageBox msgBox(this);
    msgBox.setText(text);
    msgBox.exec();
}


////////////////////////////////////////////////////////////////////////////////
// Terminate
//
//
//
////////////////////////////////////////////////////////////////////////////////
void MultiCamera::Terminate()
{
    CAM_API_STATUS status;

    // Set terminating image thread flag.
    if(this->m_hImgUpdThrd != NULL)
    {
        this->m_hImgUpdThrd->m_bTerminate = true;
    }

    // Send Acquisition stop command to camera,
    StopAcquisition();

    TerminateImgThrd();

    CloseStream();

    //
    CloseAllCameras();

    // Terminate TeliCamApi
    status = Sys_Terminate();
    if( status != CAM_API_STS_SUCCESS)
    {
        SetError(QString("Sys_Terminate error!(%1)").arg(status));
        return;
    }
}


void MultiCamera::UpdateCameraInfo()
{
    CAM_API_STATUS status;
    CAM_INFO sCameraInfo;

    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        // Get camera information
        status = Cam_GetInformation((CAM_HANDLE)NULL, i, &sCameraInfo);
        if( status != CAM_API_STS_SUCCESS)
        {	//Failed to get camera information.
            SetError(QString("Cam_GetInformation error!(%1)").arg(status));
            break;
        }

        // show camera information
        this->pArrayGroupBox[i]->setTitle(QString("ModelName : %1 , SerialNo : %2").arg(sCameraInfo.szModelName).arg(sCameraInfo.szSerialNumber));

        ui->tabWidget->setTabText(i, QString("%1(%2)").arg(sCameraInfo.szModelName).arg(sCameraInfo.szSerialNumber));
        ui->tabWidget->setTabEnabled(i, true);
    }
}


/////////////////////////////////////////////////////////
// Get Gain feature. (*All camera)
//
void MultiCamera::UpdateGain()
{
    CAM_API_STATUS status;
    float64_t dGainMin;
    float64_t dGainMax;
    float64_t dGain;

    for (uint32_t camNo = 0; camNo < MAX_CAMERA; camNo++)
    {
        if (camNo < this->m_uiCamCount)
        {
            status = GetCamGainMinMax(this->m_hCam[camNo], &dGainMin, &dGainMax);
            if( status == CAM_API_STS_SUCCESS)
            {
                this->pArrayDoubleSpinBoxGain[camNo]->setMinimum(dGainMin);
                this->pArrayDoubleSpinBoxGain[camNo]->setMaximum(dGainMax);
            }
            status = GetCamGain(this->m_hCam[camNo], &dGain);
            if( status == CAM_API_STS_SUCCESS)

            {
                this->pArrayDoubleSpinBoxGain[camNo]->setValue(dGain);
            }

            this->pArrayDoubleSpinBoxGain[camNo]->setEnabled(true);
        }
        else
        {
            this->pArrayDoubleSpinBoxGain[camNo]->setEnabled(false);
        }
    }
}

/////////////////////////////////////////////////////////
// Get Framerate feature. (*All camera)
//
void MultiCamera::UpdateFrameRate()
{
    CAM_API_STATUS status;
    float64_t dFrameRateMin;
    float64_t dFrameRateMax;
    float64_t dFrameRate;

    for (uint32_t camNo = 0; camNo < MAX_CAMERA; camNo++)
    {
        if ((!this->m_bAcquisActive) && (camNo < this->m_uiCamCount))
        {
            status = GetCamAcquisitionFrameRateMinMax(this->m_hCam[camNo], &dFrameRateMin, &dFrameRateMax);
            if( status == CAM_API_STS_SUCCESS)
            {
                this->pArrayDoubleSpinBoxFrameRate[camNo]->setMinimum(dFrameRateMin);
                this->pArrayDoubleSpinBoxFrameRate[camNo]->setMaximum(dFrameRateMax);
            }
            status = GetCamAcquisitionFrameRate(this->m_hCam[camNo], &dFrameRate);
            if( status == CAM_API_STS_SUCCESS)
            {
                this->pArrayDoubleSpinBoxFrameRate[camNo]->setValue(dFrameRate);
            }

            this->pArrayDoubleSpinBoxFrameRate[camNo]->setEnabled(true);
        }
        else
        {
            this->pArrayDoubleSpinBoxFrameRate[camNo]->setEnabled(false);
         }
    }
}


////////////////////////////////////////////////////////////////////////////////
// Close thread
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::TerminateImgThrd()
{

    // Close thread handle of Image updating threads.
    if(this->m_hImgUpdThrd != NULL)
    {
        this->m_hImgUpdThrd->m_bTerminate = true;
        this->m_hImgUpdThrd->wait();

        delete this->m_hImgUpdThrd;

        this->m_hImgUpdThrd = NULL;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Initialize
//
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::Initialize()
{
    CAM_API_STATUS status;

    for (uint32_t camNo = 0; camNo < MAX_CAMERA; camNo++)
    {
        this->pArrayGroupBox[camNo]->setTitle("");
        this->pArrayLabelInfo[camNo]->setText("");

        ui->tabWidget->setTabEnabled(camNo, false);
    }

    // Get number of cameras
    status = Sys_GetNumOfCameras(&this->m_uiCamCount);
    if (status != CAM_API_STS_SUCCESS)
    {	// Failed to get count of cameras.
        SetError(QString("Sys_GetNumOfCameras error!(%1)").arg(status));
        return false;
    }
    // Show detected camera count.
    ui->textEditNumOfCameras->clear();
    ui->textEditNumOfCameras->insertPlainText(QString("%1").arg(this->m_uiCamCount));

    if (this->m_uiCamCount == 0)
    {
        SetError("No cameras found.");
        return false;
    }

    if (m_uiCamCount > MAX_CAMERA)
    {	// if detected too much cameras. use first MAX_CAMERA cameras.
        m_uiCamCount = MAX_CAMERA;
    }

    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        // Open camera.
        status = Cam_Open(i,&this->m_hCam[i]);
        if (status != CAM_API_STS_SUCCESS)
        {	//Failed to open camera.
            SetError(QString("Cam_Open error!(%1)").arg(status));
            return false;
        }

        // AcquisitionFrameRateControl.
        CAM_TYPE eType;
        status = GetCamTypeFromCamHandle(this->m_hCam[i], &eType);
        if (status == CAM_API_STS_SUCCESS)
        {
            if (eType == CAM_TYPE_U3V)
            {
                // Set AcquisitionFramerRateControl NoSpecify.
                status = SetCamAcquisitionFrameRateControl(this->m_hCam[i], CAM_ACQ_FRAME_RATE_CTRL_MANUAL);
                if (status != CAM_API_STS_SUCCESS)
                {
                    SetError(QString("SetCamAcquisitionFrameRateControl error!(%1)").arg(status));
                    return false;
                }
            }
        }
    }

    for (uint32_t i = 0; i < MAX_CAMERA; i++)
    {
        m_uiFrmCount[i]	= 0;
        m_uiFrmIndex[i]	= 0;
    }

    UpdateCameraInfo();
    UpdateGain();
    UpdateFrameRate();

    this->m_bIsOpen = true;

    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Close All cameras
//
//
////////////////////////////////////////////////////////////////////////////////
void MultiCamera::CloseAllCameras()
{
    for (uint32_t i = 0; i < MAX_CAMERA; i++)
    {
        if (this->m_hCam[i] != (CAM_HANDLE)NULL)
        {
            Cam_Close(this->m_hCam[i]);
            this->m_hCam[i] = (CAM_HANDLE)NULL;
        }
    }

    this->m_bIsOpen = false;
}

////////////////////////////////////////////////////////////////////////////////
// Set Free=Running mode
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::SetFreeRunningMode()
{
    CAM_API_STATUS status;

    // Set TriggerMode off (Set Free-running mode)
    bool8_t	cDat = 0; // Off (Off=Freeran)
    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        status = SetCamTriggerMode(this->m_hCam[i], cDat);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("SetCamTriggerMode error!(%1)").arg(status));
            return false;
        }
    }
//    GetTriggerTypeInfo();
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Get camera parameters
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::GetCameraParameter()
{
    CAM_API_STATUS status;
    bool res;
    uint32_t  uiData;

    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        // Get maximum image size.
        status = GetCamWidth(m_hCam[i], &uiData);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("GetCamWidth error!(%1)").arg(status));
            return false;
        }
        this->m_uiCamImgWidth[i] = uiData;

        status = GetCamHeight(m_hCam[i], &uiData);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("GetCamHeight error!(%1)").arg(status));
            return false;
        }
        this->m_uiCamImgHeight[i] = uiData;

        // Set pixel format information.
        res = SetPixelFormatInfo(i);
        if (res != true)
        {
            return false;
        }
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Open Stream channel (interface)
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::OpenStream()
{
    CAM_API_STATUS  status;
    uint32_t  uiPyld = 0;

    //
    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        // Open stream channel.
        //  Use default Payload size (pyld = 0) and  MaxPacketSize(U3v : 65536, GEV : 1500)
        //
        //  If 0x0000100C(CAM_API_STS_TOO_MANY_PACKET_MISSING) is returned from Strm_DequeueRequest() in ImgRcvThread,
        //  try enabling Jumbo-Packet feature of the network adapter and changing MaxPacketSize value of the opening stream, in the case of GigE Vision camera.
        //
        //  If Junmbo-frame value of network adapter is 9014 or 9000 or 9k, Use 9000 for MaxPacketSize.
        //  Value "9014" includes ehernet header length 14 bytes. (Source MAC address 6 bytes + Destination MAC address 6 bytes + Protocol ID 2 bytes)
        //
        //  If Junmbo-frame value of network adapter is 4088 or 4074 or 4k, Use 4074 for MaxPacketSize.
        //
        //  If too large value is set to MaxPacketSize argument, 0xFFFFFFFF (CAM_API_STS_UNSUCCESSFUL) will be returned from Strm_Open() API.
        //
        //  The following is code of opening stream with MAxPacketsIze 9000.
        //
        // m_uiStatus = Strm_OpenSimple(m_hCam[i], &m_hStrm[i], &uiPyld, NULL, 5, 9000);
        //
        //
        status = Strm_OpenSimple(this->m_hCam[i], &this->m_hStrm[i], &uiPyld, NULL);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("Strm_OpenSimple error!(%1)").arg(status));
            return false;
        }

        // Allocate memory for showing image in main form. (32bit BGRA)
        this->m_pvImgBGRA[i] = malloc(this->m_uiCamImgWidth[i] * this->m_uiCamImgHeight[i] * 4);
        if (this->m_pvImgBGRA[i] == NULL)
        {
            SetError(QString("malloc error!"));
            return false;
        }

        // Set callback for image acquired event
        status = Strm_SetCallbackImageAcquired(this->m_hStrm[i], this, &CallBackImageAcquired);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("Strm_SetCallbackImageAcquired error!(%1)").arg(status));
            return false;
        }

        // Set callback for stream error event
        status = Strm_SetCallbackImageError(this->m_hStrm[i], this, &CallBackErrorImage);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("Strm_SetCallbackImageError error!(%1)").arg(status));
            return false;
        }

        // Set callback for buffer busy
        status = Strm_SetCallbackBufferBusy(this->m_hStrm[i], this, &CallbackBufferBusy);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("Strm_SetCallbackBufferBusy error!(%1)").arg(status));
            return false;
        }
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Close Stream
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::CloseStream()
{
    for (uint32_t i = 0; i < this->m_uiCamCount; i++)
    {
        // Close stream.
        if (this->m_hStrm[i] != (CAM_STRM_HANDLE)NULL)
        {
            Strm_Close(this->m_hStrm[i]);
            this->m_hStrm[i] = (CAM_STRM_HANDLE)NULL;
        }

        // Release BGRA image data of each camera..
        if (this->m_pvImgBGRA[i] != NULL)
        {
            free(this->m_pvImgBGRA[i]);
            this->m_pvImgBGRA[i] = NULL;
        }

        this->m_uiFrmCount[i] = 0;
    }
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Start Acquisition
//
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::StartAcquisition()
{
    CAM_API_STATUS  status;

    // Create thread for updating images in main form
    if (!this->m_hImgUpdThrd)
    {
        this->m_hImgUpdThrd =  new DispThread( this );

        connect(this->m_hImgUpdThrd, SIGNAL(DispFps(int, double)), this, SLOT(disp_fps(int, double)));
        connect(this->m_hImgUpdThrd, SIGNAL(DispImage(int, QImage)), this, SLOT(disp_image(int, QImage)));
    }

    // Clear thread-closing request flag.
    this->m_hImgUpdThrd->m_bTerminate = false;

    this->m_hImgUpdThrd->start();

    // Send acquisition command to camera
    for (uint32_t i = 0; i < this->m_uiCamCount; ++i)
    {
        status = Strm_Start(this->m_hStrm[i]);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("Strm_Start error!(%1)").arg(status));
            return false;
        }
    }

    // Set Acquisition active flag.
    this->m_bAcquisActive = true;

    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Stop Acquisition
//
// Send StopAcquisition command to camera
//
////////////////////////////////////////////////////////////////////////////////
 bool8_t MultiCamera::StopAcquisition()
 {
     CAM_API_STATUS  status;

     // Set Acquisition active flag.
     this->m_bAcquisActive = false;

     // send acquisition stop command to camera
     for (uint32_t i = 0; i < this->m_uiCamCount; ++i)
     {
         if (this->m_hStrm[i] != (CAM_STRM_HANDLE)NULL)
         {
             status = Strm_Stop(this->m_hStrm[i]);
             if (status != CAM_API_STS_SUCCESS)
             {
                 SetError(QString("Strm_Stop error!(%1)").arg(status));
                 return false;
             }
         }
     }

     return true;
 }


////////////////////////////////////////////////////////////////////////////////
// Set PixelFormat information.
//
// [im] uiIndex				index of camera that PixelFormat information should be updated.
//
// return value				if success 0
//
////////////////////////////////////////////////////////////////////////////////
bool8_t MultiCamera::SetPixelFormatInfo(uint32_t uiIndex)
{
    CAM_API_STATUS status;
    CAM_PIXEL_FORMAT uiPixFormat;

    if(uiIndex < this->m_uiCamCount)
    {
        // Get pixel format
        status = GetCamPixelFormat(m_hCam[uiIndex], &uiPixFormat);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("GetCamPixelFormat error!(%1)").arg(status));
            return false;
        }
        this->m_uiCamPixFormat[uiIndex] = uiPixFormat;
    }
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Callback for handling received image.
//
// [in] hCam					Handler of camera.
// [in] hStrm					Handler of stream.
// [in] psImageInfo				additional information of received imagee.
// [in] uiBufferIndex			Index of Buffer.
// [in] pvContext			Additional caller-specified context.
//
//
////////////////////////////////////////////////////////////////////////////////
void MultiCamera::CallBackImageAcquired(CAM_HANDLE hCam, CAM_STRM_HANDLE hStrm, CAM_IMAGE_INFO *psImageInfo, uint32_t uiBufferIndex, void *pvContext)
{
    Q_UNUSED(hStrm);
    Q_UNUSED(uiBufferIndex);

    MultiCamera *pDlg = reinterpret_cast<MultiCamera *>(pvContext);
    // Get camera index from camera handle.
    uint32_t uiIndexCam;
    CAM_API_STATUS uiSts = GetCamIndexFromCamHandle(hCam, &uiIndexCam);
    if (uiSts != CAM_API_STS_SUCCESS)
    {
        pDlg->DispInfo(uiIndexCam, tr("ImageAcquired error!"));
        return;
    }

    //***************************************************
    // Insert image processing code here.
    //
    //   Image processing which will take a long time
    //   should not be executed in this method.
    //
    //***************************************************

    if(pDlg->m_bFrmIndexValid[uiIndexCam])
    {
        pDlg->m_uiFrmCount[uiIndexCam] += ((uint32_t)(psImageInfo->ullBlockId) - pDlg->m_uiFrmIndex[uiIndexCam]);
        pDlg->m_uiFrmIndex[uiIndexCam] = static_cast<uint32_t>(psImageInfo->ullBlockId);
    }
    else
    {
        pDlg->m_bFrmIndexValid[uiIndexCam] = true;
        pDlg->m_uiFrmCount[uiIndexCam]++ ;
        pDlg->m_uiFrmIndex[uiIndexCam] = static_cast<uint32_t>(psImageInfo->ullBlockId);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Callback for handling receive error.
//
// [in] hCam					Handler of camera.
// [in] hStrm					Handler of stream.
// [in] uiErrorStatus			Error status.
// [in] uiBufferIndex			Index of buffer.
// [in] pvContext				Additional caller-specified context.
//
//
////////////////////////////////////////////////////////////////////////////////
void MultiCamera::CallBackErrorImage(CAM_HANDLE hCam, CAM_STRM_HANDLE hStrm, CAM_API_STATUS uiErrorStatus, uint32_t uiBufferIndex, void  *pvContext)
{
    Q_UNUSED(hStrm);
    Q_UNUSED(uiErrorStatus);
    Q_UNUSED(uiBufferIndex);

    MultiCamera *pDlg = reinterpret_cast<MultiCamera *>(pvContext);

    if (pDlg->m_bAcquisActive)
    {
        // Get camera index from camera handle.
        uint32_t uiIndexCam;
        CAM_API_STATUS uiSts = GetCamIndexFromCamHandle(hCam, &uiIndexCam);
        if (uiSts == CAM_API_STS_SUCCESS)
        {
            pDlg->DispInfo(uiIndexCam, tr("Error image!"));
        }
    }
}

void MultiCamera::CallbackBufferBusy(CAM_HANDLE hCam, CAM_STRM_HANDLE hStrm, uint32_t uiBufferIndex, void* pvContext)
{
    Q_UNUSED(hStrm);
    Q_UNUSED(uiBufferIndex);

    MultiCamera *pDlg = reinterpret_cast<MultiCamera *>(pvContext);
    // Get camera index from camera handle.
    uint32_t uiIndexCam;
    CAM_API_STATUS uiSts = GetCamIndexFromCamHandle(hCam, &uiIndexCam);
    if (uiSts == CAM_API_STS_SUCCESS)
    {
        pDlg->DispInfo(uiIndexCam, tr("Buffer busy!"));
    }
}

void MultiCamera::SetGain(uint32_t camNo, double value)
{
    if (this->m_bIsOpen)
    {
        CAM_API_STATUS  status;

        status = SetCamGain(this->m_hCam[camNo], (float64_t)value);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("SetCamGain error!(%1)").arg(status));
        }
    }
}

void MultiCamera::SetFramerate(uint32_t camNo, double value)
{
    if (this->m_bIsOpen)
    {
        CAM_API_STATUS  status;

        status = SetCamAcquisitionFrameRate(this->m_hCam[camNo], (float64_t)value);
        if (status != CAM_API_STS_SUCCESS)
        {
            SetError(QString("SetCamAcquisitionFrameRate error!(%1)").arg(status));
        }

         float64_t dFrameRate;
        status = GetCamAcquisitionFrameRate(this->m_hCam[camNo], &dFrameRate);
        if( status == CAM_API_STS_SUCCESS)
        {
            this->pArrayDoubleSpinBoxFrameRate[camNo]->setValue(dFrameRate);
        }
    }
}

void MultiCamera::DispInfo(uint32_t camNo, QString message)
{
    this->pArrayLabelInfo[camNo]->setText(message);
}

//////////////////////////////////////////////////////////////////////////////////

void MultiCamera::disp_image(int camNo, QImage pimage)
{
    if (this->m_bAcquisActive && (m_uiSelectedTabIndex == (uint32_t)camNo))
    {
        this->pArrayLabelImage[camNo]->setPixmap(QPixmap::fromImage(pimage).scaled(this->pArrayLabelImage[camNo]->width(), this->pArrayLabelImage[camNo]->height()));
        this->pArrayLabelImage[camNo]->update();
    }
}

void MultiCamera::disp_fps(int camNo, double fps)
{
    this->pArrayLabelFps[camNo]->setText(QString("fps : %1").arg(fps, 7, 'f', 3));
    this->pArrayLabelFps[camNo]->update();
}

//////////////////////////////////////////////////////////////////////////////////

void MultiCamera::on_pushButtonUpdateCameraList_clicked()
{
    bool8_t res;

    ui->pushButtonUpdateCameraList->setEnabled(false);
    ui->pushButtonStartAcquisition->setEnabled(false);

    do
    {
        // Stop receiving image
        CloseAllCameras();

        //
        res = Initialize();
        if (res != true)
        {
            CloseAllCameras();
            break;
        }

        //
        res = GetCameraParameter();
        if (res != true)
        {
            CloseAllCameras();
            break;
        }

        //
        res = SetFreeRunningMode();
        if (res != true)
        {
            CloseAllCameras();
            break;
        }
    } while(false);


    ui->pushButtonUpdateCameraList->setEnabled(true);
    if (this->m_uiCamCount > 0)
    {
        ui->pushButtonStartAcquisition->setEnabled(true);
    }
}

void MultiCamera::on_pushButtonStartAcquisition_clicked()
{
    bool res;

    // Prepare for receiving image stream from camera.
    //  *Create stream-receiving complete event and open stream channel.
    res = OpenStream();
    if (res != true)
    {
        CloseAllCameras();
        return;
    }

    // Start sending image from camera.
    //  * Send StartAcquisition command to camera.
    res = StartAcquisition();
    if (res != true)
    {
        CloseAllCameras();
        return;
    }

    ui->pushButtonUpdateCameraList->setEnabled(false);
    ui->pushButtonStartAcquisition->setEnabled(false);
    ui->pushButtonStopAcquisition->setEnabled(true);
    ui->doubleSpinBoxFrameRate1->setEnabled(false);
    ui->doubleSpinBoxFrameRate2->setEnabled(false);
    ui->doubleSpinBoxFrameRate3->setEnabled(false);
    ui->doubleSpinBoxFrameRate4->setEnabled(false);
}

void MultiCamera::on_pushButtonStopAcquisition_clicked()
{
    // Set terminating image thread flag.
    if(this->m_hImgUpdThrd != NULL)
    {
        this->m_hImgUpdThrd->m_bTerminate = true;
    }

    // Send Acquisition stop command to camera,
    StopAcquisition();

    TerminateImgThrd();

    CloseStream();

    ui->pushButtonUpdateCameraList->setEnabled(true);
    ui->pushButtonStartAcquisition->setEnabled(true);
    ui->pushButtonStopAcquisition->setEnabled(false);
    ui->doubleSpinBoxFrameRate1->setEnabled(true);
    ui->doubleSpinBoxFrameRate2->setEnabled(true);
    ui->doubleSpinBoxFrameRate3->setEnabled(true);
    ui->doubleSpinBoxFrameRate4->setEnabled(true);
}

void MultiCamera::on_doubleSpinBoxGain1_valueChanged(double arg1)
{
    SetGain(0, arg1);
}

void MultiCamera::on_doubleSpinBoxGain2_valueChanged(double arg1)
{
    SetGain(1, arg1);
}

void MultiCamera::on_doubleSpinBoxGain3_valueChanged(double arg1)
{
    SetGain(2, arg1);
}

void MultiCamera::on_doubleSpinBoxGain4_valueChanged(double arg1)
{
    SetGain(3, arg1);
}

void MultiCamera::on_doubleSpinBoxFrameRate1_valueChanged(double arg1)
{
    SetFramerate(0, arg1);
}

void MultiCamera::on_doubleSpinBoxFrameRate2_valueChanged(double arg1)
{
    SetFramerate(1, arg1);
}

void MultiCamera::on_doubleSpinBoxFrameRate3_valueChanged(double arg1)
{
    SetFramerate(2, arg1);
}

void MultiCamera::on_doubleSpinBoxFrameRate4_valueChanged(double arg1)
{
    SetFramerate(3, arg1);
}

void MultiCamera::on_tabWidget_currentChanged(int index)
{
    m_uiSelectedTabIndex = index;
}
