/******************************************************************************/
/* GrabStream_RegBU.cpp                                                       */
/*							                                                  */
/* TeliCamSDK sample console application for BU series camera.                */
/*                                                                            */
/* Access camera feature registers directly as independent registers,  		  */
/* instead of using TeliCamApi functions which call GenApi.                   */
/*									                                          */
/******************************************************************************/

#if defined (_WIN32)
	#include <windows.h>
	#include <tchar.h>
	#include <conio.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>

#include "TeliCamApi.h"
#include "TeliCamUtl.h"
#include "RegisterMap_BU.h"

using namespace std;
using namespace Teli;



/******************************************************************************/
/* Prototype declares                                                         */
/******************************************************************************/
uint32_t Initialize();
void Terminate();

uint32_t SetFeature();
uint32_t SetSoftTriggerMode();
uint32_t OpenStream();
uint32_t CloseStream();

uint32_t StartAcquisition();
uint32_t StopAcquisition();

uint32_t ProcessLoop();
uint32_t DispFeature();
uint32_t ImageHandling();
uint32_t SetExposureTime();
uint32_t SetGain();

/*****************************************************************************/
/* Static Values                                                             */
/*****************************************************************************/

//
static CAM_API_STATUS  s_uiStatus = CAM_API_STS_SUCCESS;

static bool8_t s_bU3vCamera;

// Position, size, pixel coding, pixel size of image
static const int32_t	s_uiOfstX		= 0;		// Offset X
static const int32_t	s_uiOfstY		= 0;		// Offset Y
static       uint32_t	s_uiWidth		= 0;		// Width
static       uint32_t	s_uiHeight		= 0;		// Height
static const uint32_t	s_uiPxlCod		= RegMapBU::RV_IMGFMT_MONO;  // Only Monochrome is supported in this sample
static const uint32_t	s_uiPxlSiz		= RegMapBU::RV_IMGFMT_BPP8;  // Only 8bit is supported in this sample
static		 uint8_t	*s_pucImgBuf	= NULL;		// Local image buffer for calculating average of received image pixel
static       uint32_t	s_uiImgBufSize	= 0;		// Size of image buffer.

// Exposure, gain, framer rate.
static double			s_dExpBase		= 0.0;
static uint32_t			s_uiExpMin		= 0;
static uint32_t			s_uiExpMax		= 0;
static double			s_dGainBase		= 0.0;
static uint32_t			s_uiGainMin		= 0;
static uint32_t			s_uiGainMax		= 0;
static double			s_dFrateBase	= 0.0;
static double			s_dMultDspExp	= 1000;

// Handles
static CAM_HANDLE		s_hCam			= (CAM_HANDLE)NULL;		// Camera handle
static CAM_STRM_HANDLE	s_hStrm			= (CAM_STRM_HANDLE)NULL;		// Stream handle

#if defined (_WIN32)
static HANDLE			s_hStrmEvt		= NULL;		// Completion event for stream.
#else
static SIGNAL_HANDLE	s_hStrmEvt;
#endif

/*****************************************************************************/
/* Functions                                                                 */
/*****************************************************************************/


#if defined (_WIN32)
int _tmain(int argc, _TCHAR* argv[])
#else
int main()
#endif
{
	uint32_t  uiStatus = 0;

	printf("GrabStreamSimpleRegBU sample application for BUs and TeliCamSDK\n\n");
	printf("Copyrights (C) 2015 TOSHIBA TELI CORPORATION. All rights reserved.\n\n");
	printf(" Access camera feature registers directly as independent registers,\n");
	printf(" instead of using TeliCamApi functions which call GenApi.\n\n");

	// Initialize.
	do
	{
		uiStatus = Initialize();
		if (uiStatus != 0)
		{
			printf("Failed Initialize(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}


		// Set features.
		uiStatus = SetFeature();
		if (uiStatus != 0)
		{
			printf("Failed SetFeature(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}

		// Set software one-shot trigger mode.
		uiStatus = SetSoftTriggerMode();
		if (uiStatus != 0)
		{
			printf("Failed SetTriggerMode(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}

		// Open stream.
		uiStatus = OpenStream();
		if (uiStatus != 0)
		{
			printf("Failed OpenStream(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}

		// Start acquisition
		uiStatus = StartAcquisition();
		if (uiStatus != 0)
		{
			printf("Failed StartAcquisition(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}

		// Receive images.
		uiStatus = ProcessLoop();
		if (uiStatus != 0)
		{
			printf("Failed ProcessLoop(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
		}

		// Stop acquisition.
		uiStatus = StopAcquisition();
		if(uiStatus != 0)
		{
			printf("Failed StopAcquisition(), Location = %d, Status = 0x%08X.\n", uiStatus, s_uiStatus);
			break;
		}
	} while(0);

	// Close stream.
	CloseStream();

	// Terminate.
	Terminate();

	printf("Finish!\n\n");
	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t Initialize()
{
	uint32_t  uiNum;
	uint32_t  uiMax;


	printf("\n");
	printf("Initialize() started!\n");

	// API initialization.
	s_uiStatus = Sys_Initialize();
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 1;
	}

	// Get SDK system information.
	CAM_SYSTEM_INFO sSysInfo;
	s_uiStatus = Sys_GetInformation(&sSysInfo);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 2;
	}
	printf(" <System information>\n");
	printf("  TeliU3vDriver  version : %s\n", sSysInfo.sU3vInfo.szDriverVersion);
	printf("  TeliU3vApi2    version : %s\n", sSysInfo.sU3vInfo.szDllVersion);
	printf("  TeliU3vCamApi  version : %s\n", sSysInfo.sU3vInfo.szDllExVersion);
	printf("  TeliCamApi     version : %s\n", sSysInfo.szDllVersion);

	// Get uiNumber of camera.
	s_uiStatus = Sys_GetNumOfCameras(&uiNum);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 3;
	}
	if (uiNum == 0)
	{
 		printf(" No cameras found.\n");
		return 4;
	}
	printf(" %d camera(s) found.\n", uiNum);

	// Get camera information.
	CAM_INFO		sCamInfo;
	U3V_CAM_INFO	*psU3vCamInfo;
	GEV_CAM_INFO	*psGevCamInfo;

	int iIndexU3v = -1;						// index of USB3 Vision camera

	for (int i = 0; i < (int)uiNum; i++)
	{
		memset((void*)&sCamInfo, 0, sizeof(CAM_INFO));
		// Get information of a camera.
		s_uiStatus = Cam_GetInformation((CAM_HANDLE)NULL, i, &sCamInfo);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 5;
		}

		printf("\n<Camera%d information>\n", i);
		if (sCamInfo.eCamType == CAM_TYPE_U3V)
		{
			printf(" Type                               : USB3 Vision camera\n");
			if(iIndexU3v < 0)
			{
				iIndexU3v =i;
			}
		}
		else if (sCamInfo.eCamType == CAM_TYPE_GEV)
			printf(" Type                               : GigE Vision camera\n");

		printf(" Manufacturer                       : %s\n", sCamInfo.szManufacturer);
		printf(" Model name                         : %s\n", sCamInfo.szModelName);
		printf(" Serial number                      : %s\n", sCamInfo.szSerialNumber);
		printf(" User defined name                  : %s\n", sCamInfo.szUserDefinedName);

		if (sCamInfo.eCamType == CAM_TYPE_U3V) {
			psU3vCamInfo = &sCamInfo.sU3vCamInfo;

			printf(" U3v family name                    : %s\n", psU3vCamInfo->szFamilyName);
			printf(" U3v device version                 : %s\n", psU3vCamInfo->szDeviceVersion);
			printf(" U3v manufacturer information       : %s\n", psU3vCamInfo->szManufacturerInfo);
			printf(" U3v adapter vendor ID              : 0x%04X\n", psU3vCamInfo->uiAdapterVendorId);
			printf(" U3v adapter device ID              : 0x%04X\n", psU3vCamInfo->uiAdapterDeviceId);
			printf(" U3v Adapter default MaxPacketSize  : %d\n", psU3vCamInfo->uiAdapterDfltMaxPacketSize);
		} else if (sCamInfo.eCamType == CAM_TYPE_GEV) {
			psGevCamInfo = &sCamInfo.sGevCamInfo;

			printf(" Gev display name                   : %s\n", psGevCamInfo->szDisplayName);
			printf(" Gev MAC address                    : %02X-%02X-%02X-%02X-%02X-%02X\n",
				psGevCamInfo->aucMACAddress[0],
				psGevCamInfo->aucMACAddress[1],
				psGevCamInfo->aucMACAddress[2],
				psGevCamInfo->aucMACAddress[3],
				psGevCamInfo->aucMACAddress[4],
				psGevCamInfo->aucMACAddress[5]);
			printf(" Gev support IP LLA                 : %d\n", psGevCamInfo->cSupportIP_LLA);
			printf(" Gev support IP DHCP                : %d\n", psGevCamInfo->cSupportIP_DHCP);
			printf(" Gev Support IP Persistent-IP       : %d\n", psGevCamInfo->cSupportIP_Persistent);
			printf(" Gev current IP LLA                 : %d\n", psGevCamInfo->cCurrentIP_LLA);
			printf(" Gev current IP DHCP                : %d\n", psGevCamInfo->cCurrentIP_DHCP);
			printf(" Gev current IP Persistent-IP       : %d\n", psGevCamInfo->cCurrentIP_Persistent);
			printf(" Gev IP Address                     : %d.%d.%d.%d\n",
				psGevCamInfo->aucIPAddress[0],
				psGevCamInfo->aucIPAddress[1],
				psGevCamInfo->aucIPAddress[2],
				psGevCamInfo->aucIPAddress[3]);
			printf(" Gev subnet mask                    : %d.%d.%d.%d\n",
				psGevCamInfo->aucSubnet[0],
				psGevCamInfo->aucSubnet[1],
				psGevCamInfo->aucSubnet[2],
				psGevCamInfo->aucSubnet[3]);
			printf(" Gev default gateway                : %d.%d.%d.%d\n",
				psGevCamInfo->aucGateway[0],
				psGevCamInfo->aucGateway[1],
				psGevCamInfo->aucGateway[2],
				psGevCamInfo->aucGateway[3]);
			printf(" Gev adapter MAC address            : %02X-%02X-%02X-%02X-%02X-%02X\n",
				psGevCamInfo->aucAdapterMACAddress[0],
				psGevCamInfo->aucAdapterMACAddress[1],
				psGevCamInfo->aucAdapterMACAddress[2],
				psGevCamInfo->aucAdapterMACAddress[3],
				psGevCamInfo->aucAdapterMACAddress[4],
				psGevCamInfo->aucAdapterMACAddress[5]);
			printf(" Gev adapter IP address             : %d.%d.%d.%d\n",
				psGevCamInfo->aucAdapterIPAddress[0],
				psGevCamInfo->aucAdapterIPAddress[1],
				psGevCamInfo->aucAdapterIPAddress[2],
				psGevCamInfo->aucAdapterIPAddress[3]);
			printf(" Gev adapter subnet mask            : %d.%d.%d.%d\n",
				psGevCamInfo->aucAdapterSubnet[0],
				psGevCamInfo->aucAdapterSubnet[1],
				psGevCamInfo->aucAdapterSubnet[2],
				psGevCamInfo->aucAdapterSubnet[3]);
			printf(" Gev adapter default gateway        : %d.%d.%d.%d\n",
				psGevCamInfo->aucAdapterGateway[0],
				psGevCamInfo->aucAdapterGateway[1],
				psGevCamInfo->aucAdapterGateway[2],
				psGevCamInfo->aucAdapterGateway[3]);
			printf(" Gev adapter display name           : %s\n", psGevCamInfo->szAdapterDisplayName);
		}
	}

	if(iIndexU3v >= 0)
	{
		// Open USB3camera that is detected first, in this sample code.
		// GenICam camera description file will not be loaded.
		s_uiStatus = Cam_Open(static_cast<uint32_t>(iIndexU3v), &s_hCam, NULL, false);

		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 6;
		}
		printf(" Camera opened.\n");

		// Check camera interface type.
		CAM_TYPE eType;
		s_uiStatus = GetCamTypeFromCamHandle(s_hCam, &eType);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 7;
		}

		if (eType == CAM_TYPE_U3V)
		{
			s_bU3vCamera = true;
		} else
		{
			s_bU3vCamera = false;
		}

		// Get sensor width.
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_IMGFMT_IMAGE_TOTAL_SIZE_X_I,1, &uiMax);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 8;
		}
		printf(" Pixel Width                        : %4d\n",uiMax);

		// Get sensor height.
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_IMGFMT_IMAGE_TOTAL_SIZE_Y_I,1, &uiMax);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 9;
		}
		printf(" Pixel Height                       : %4d\n",uiMax);

		printf("Initialize() successful.\n");
	}
	else
	{
		printf("\nNo USB3 Vision camera found!\n\n");
		return 8;
	}

	return 0;
}

/////////////////////////////////////////////////////////
//
void Terminate()
{
	printf("\n");
	printf("Terminate() started!\n");

	// Close camera.
	if (s_hCam != (CAM_HANDLE)NULL)
	{
		Cam_Close(s_hCam);
		s_hCam = (CAM_HANDLE)NULL;
	}

	// Terminate system.
	Sys_Terminate();

	printf("Terminate() successful.\n");
}

/////////////////////////////////////////////////////////
//
uint32_t SetFeature()
{
	uint32_t  uiData, uiMult, uiDiv, uiMax, uiMin;


	printf("\n");
	printf("SetFeature() started!\n");

	// ExposureTime / ExposureMode.
	{
		// Set ExposureTimeControl "Manual", in this sample code.
		uiData = RegMapBU::FEATCNT_MANUAL;
		s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_CNTRL_I,
								1, &uiData);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 20;
		}

		// In Exposure FeatureCSR, actual value can be calculated using normal multiply and divide calculation.
		//
		// Get ExposureTimeMult
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_INT32B + RegMapBU::OFS_FCSR_INT32_MULT_I,
								1, &uiMult);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 21;
		}

		// Get ExposureTimeDiv
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_INT32B + RegMapBU::OFS_FCSR_INT32_DIV_I,
								1, &uiDiv);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 22;
		}

		// Get ExposureTimeMin
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_INT32B + RegMapBU::OFS_FCSR_INT32_MIN_I,
								1, &s_uiExpMin);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 23;
		}

		// Get ExposureTimeMax
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_INT32B + RegMapBU::OFS_FCSR_INT32_MAX_I,
								1, &s_uiExpMax);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 24;
		}

		s_dExpBase = (double)uiMult / (double)uiDiv;
	}

	// GainAuto.
	{
		// In Gain FeatureCSR, actual value can be calculated using normal multiply and divide calculation.
		//
		// Get GainMult
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_INT32B + RegMapBU::OFS_FCSR_INT32_MULT_I,
								1, &uiMult);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 26;
		}

		// Get GainDiv
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_INT32B + RegMapBU::OFS_FCSR_INT32_DIV_I,
								1, &uiDiv);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 27;
		}

		// Get GainMin
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_INT32B + RegMapBU::OFS_FCSR_INT32_MIN_I,
								1, &s_uiGainMin);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 28;
		}

		// Get GainMax
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_INT32B + RegMapBU::OFS_FCSR_INT32_MAX_I,
								1, &s_uiGainMax);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 29;
		}

		s_dGainBase = (double)uiMult / (double)uiDiv;
	}

	// AcquisitionFrameRateControl.
	{
		// Set AcquisitionFrameRateControl
		uiData = RegMapBU::FEATCNT_NOSPECIFY;
		s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_ACQ_FRAME_RATE_CNTRL_I,
								1, &uiData);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 30;
		}

		// In AcquisitionFrameRateMult FeatureCSR, actual value can be calculated using normal multiply and divide calculation.
		//
		// Get AcquisitionFrameRateMult
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_ACQ_FRAME_RATE_INT32B + RegMapBU::OFS_FCSR_INT32_MULT_I,
								1, &uiMult);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 31;
		}

		// Get AcquisitionFrameRateDiv
		s_uiStatus = Cam_ReadReg(s_hCam,  RegMapBU::ADR_ACQ_FRAME_RATE_INT32B + RegMapBU::OFS_FCSR_INT32_DIV_I,
								1, &uiDiv);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 32;
		}

		// Get AcquisitionFrameRateMin
		s_uiStatus = Cam_ReadReg(s_hCam,  RegMapBU::ADR_ACQ_FRAME_RATE_INT32B + RegMapBU::OFS_FCSR_INT32_MIN_I,
								1, &uiMin);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 33;
		}

		// Get AcquisitionFrameRateMax
		s_uiStatus = Cam_ReadReg(s_hCam,  RegMapBU::ADR_ACQ_FRAME_RATE_INT32B + RegMapBU::OFS_FCSR_INT32_MAX_I,
								1, &uiMax);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 34;
		}

		s_dFrateBase = (double)uiMult / (double)uiDiv; 	}
	printf("SetFeature() successful.\n");
	return 0;
}

/////////////////////////////////////////////////////////
// Trigger mode value should be set before calling Strm_StartingAcquisition().
//
uint32_t SetSoftTriggerMode()
{
// This application uses software one shot trigger mode.
//   TriggerMode : On.
//   TriggerSequence : TriggerSequence0 (External edge mode). This feature does not exist in GigE Vision camera.
//   TriggerSource	 : Software.
//
// If user application uses Nd_GetEuiNumIntValue() or Nd_SetEuiNumIntValue() for TriggerSource,
// be careful that integer value of TriggerSource is different from that of CAM_TRIGGER_SOURCE_TYPE in GigE Vision camera.
//
// Using GetCamTriggerSource(), SetCamTriggerSource, Nd_GetEuiNumStrValue() or Nd_SetEuiNumStrValue() is recommendable.
//
//            CAM_TRIGGER_SOURCE_TYPE     USB3 Vision     GigE Vision
//    Line0	            0                     0               1
//    Software         64                    64               0
//
	uint32_t  uiData;


	printf("\n");
	printf("SetTriggerMode() started!\n");

	// Set TriggerMode true, in this sample code.
	{
		uiData = RegMapBU::RV_CATC_ONOFF_ON;
		s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_TRG_TRG_MODE_I, 1, &uiData);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 40;
		}
	}

	// Set TriggerSequence TriggerSequence0.
	{
		// Trigger sequence must be set to TriggerSequence0 in software trigger mode.
		uiData = RegMapBU::RV_TRIGGER_SEQUENCE_0; // Edge mode
		s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_TRG_TRG_SEQUENCE_I, 1, &uiData);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 42;
		}
	}


	// Set TriggerSource software.
	{
		uiData = RegMapBU::RV_TRG_SRC_SOFTWARE;
		s_uiStatus = Cam_WriteReg(s_hCam,  RegMapBU::ADR_TRG_TRG_SOURCE_I, 1, &uiData);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 45;
		}
	}

	printf("SetTriggerMode() successful.\n");
	return 0;
}


/////////////////////////////////////////////////////////
//
uint32_t OpenStream()
{
	printf("\n");
	printf("OpenStream() started!\n");

	// Create completion event for stream
#if defined (_WIN32)
	s_hStrmEvt = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (s_hStrmEvt == NULL)
	{
		return 80;
	}
#else
	Sys_CreateSignal(&s_hStrmEvt);
#endif

	// Open stream channel
	// Use default MaxPacketSize(U3v : 65536 or 524288 , GEV : 1500).
	// Payload size will be written to uiPyld,
	 s_uiStatus = Strm_OpenSimple(s_hCam, &s_hStrm, &s_uiImgBufSize, s_hStrmEvt);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 81;
	}
	printf(" Payload size : %d[byte]\n", s_uiImgBufSize);

	// Allocate image bufer for receiving image data from TeliCamAPI.
#if defined (_WIN32)
	s_pucImgBuf = (uint8_t *)VirtualAlloc(NULL,
									 s_uiImgBufSize,
									 MEM_RESERVE | MEM_COMMIT,
									 PAGE_EXECUTE_READWRITE);
#else
	s_pucImgBuf = (uint8_t *)malloc( s_uiImgBufSize);
#endif

	if (s_pucImgBuf == NULL)
	{
		return 82;
	}

	printf("OpenStream() successful.\n");
	return 0;
}



/////////////////////////////////////////////////////////
//
uint32_t CloseStream()
{
	printf("\n");
	printf("CloseStream() started!\n");

	// Close stream.
	if (s_hStrm != (CAM_STRM_HANDLE)NULL)
	{
		Strm_Close(s_hStrm);
		s_hStrm = (CAM_STRM_HANDLE)NULL;
	}

	// Close completion event for stream.
	if (s_hStrmEvt != NULL)
	{
#if defined (_WIN32)
		CloseHandle(s_hStrmEvt);
#else
		Sys_CloseSignal(s_hStrmEvt);
#endif
		s_hStrmEvt = NULL;
	}

	// Release image buffer.
	if (s_pucImgBuf != NULL)
	{
#if defined (_WIN32)
		VirtualFree(s_pucImgBuf, 0, MEM_RELEASE);
#else
		free(s_pucImgBuf);
#endif
		s_pucImgBuf = NULL;
	}

	printf("CloseStream() successful.\n");
	return 0;
}



/////////////////////////////////////////////////////////
//
uint32_t StartAcquisition()
{
	printf("\n");
	printf("StartAcquisition() started!\n");

	// Enable streaming
	uint32_t uiData = RegMapBU::RV_SI_CONTROL_ENABLE;
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_SIRM_SI_CONTROL_C, 1, &uiData);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 100;
	}

	// Start Acquisition
	uiData = RegMapBU::RV_ACQ_ACQUIS_CMD_CONTINUOUS;
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_ACQ_ACQUISITION_I, 1, &uiData);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 101;
	}

	printf("StartAcquisition() successful.\n");
	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t StopAcquisition()
{
	printf("\n");
	printf("StopAcquisition() started!\n");

	// Disable streaming
	uint32_t uiData = RegMapBU::RV_SI_CONTROL_DISABLE;
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_SIRM_SI_CONTROL_C, 1, &uiData);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 120;
	}

	// Stop Acquisition
	uiData = RegMapBU::RV_ACQ_ACQUIS_CMD_STOP;
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_ACQ_ACQUISITION_I, 1, &uiData);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 121;
	}

	printf("StopAcquisition() successful.\n");
	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t SetExposureTime()
{
	uint32_t  uiExpRaw;
	double    dExp, dExpMin, dExpMax;

	// Calc min/max abs
	dExpMin = s_dExpBase * (double)s_uiExpMin;
	dExpMax = s_dExpBase * (double)s_uiExpMax;

	//
	printf(" Exposure time min. : %.4lf, max. : %.4lf[ms] >>> ", dExpMin * s_dMultDspExp, dExpMax * s_dMultDspExp);
#if defined (_WIN32)
	scanf_s("%lf", &dExp);
#else
	if (scanf("%lf", &dExp) == EOF)
	{
		return 140;
	}
#endif

	// Convert abs to raw. (In Exposure FeatureCSR, actual value can be calculated using normal multiply and divide calculation.)
	uiExpRaw = (uint32_t)(dExp / s_dMultDspExp / s_dExpBase);
	uiExpRaw = max(s_uiExpMin, min(uiExpRaw, s_uiExpMax));

	// Set ExposureTime
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_RAW_I, 1, &uiExpRaw);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 141;
	}

	// Get updated ExposureTime
	s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_RAW_I, 1, &uiExpRaw);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 142;
	}
	printf(" Current exposure time : %.4lf[ms].\n",  s_dExpBase * (double)uiExpRaw * s_dMultDspExp);

	printf("SetExposureTime() finished!\n");
	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t SetGain()
{
	uint32_t  uiGainRaw;
	double    dGain, dGainMin, dGainMax;


	printf("\n");
	printf("SetGain() started!\n");

	// (In Gain FeatureCSR, actual value can be calculated using normal multiply and divide calculation.)
	// Calc min/max abs
	dGainMin = s_dGainBase * (double)s_uiGainMin;
	dGainMax = s_dGainBase * (double)s_uiGainMax;

	printf(" Gain min. :%.4lf, max. :%.4lf [db] >>> ", dGainMin, dGainMax );
#if defined (_WIN32)
	scanf_s("%lf", &dGain);
#else
	if (scanf("%lf", &dGain) == EOF)
	{
		return 160;
	}
#endif

	// Convert abs to raw.
	uiGainRaw = (uint32_t)(dGain / s_dGainBase);
	uiGainRaw = max(s_uiGainMin, min(uiGainRaw, s_uiGainMax));

	// Set Gain
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_LMN_GAIN_RAW_I, 1, &uiGainRaw);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 161;
	}

	// Get updated Gain
	s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_RAW_I, 1, &uiGainRaw);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 162;
	}

	printf(" Current dGain : %.4lf[db].\n", s_dGainBase * (double)uiGainRaw);

	printf("SetGain() finished!\n");
	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t ImageHandling()
{
	const uint32_t	uiTimeout = 5000;	// 5sec.
	uint32_t		uiRes, uiData;
	uint32_t		uiSize  = s_uiImgBufSize;
	CAM_IMAGE_INFO	imageInfo;

	// Send Software Trigger command.
	uiData = RegMapBU::RV_TRG_SOFTWARE_IMPULSE;
	s_uiStatus = Cam_WriteReg(s_hCam, RegMapBU::ADR_TRG_SOFT_TRG_I, 1, &uiData);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 181;
	}

	// Wait for receiving image completion event.
#if defined (_WIN32)
	uiRes = WaitForSingleObject(s_hStrmEvt, uiTimeout);
	if (uiRes != WAIT_OBJECT_0)
	{	// Timeout.
		return 182;
	}
#else
	s_uiStatus = Sys_WaitForSignal(s_hStrmEvt, uiTimeout);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 182;
	}
#endif

	// Copy received image data to local buffer, and get size of received image. In this samplecode.
	s_uiStatus = Strm_ReadCurrentImage(s_hStrm, s_pucImgBuf, &uiSize, &imageInfo);
	if (s_uiStatus != CAM_API_STS_SUCCESS)
	{
		return 183;
	}
#if defined (_WIN32)
	printf("Received ImageAcquired event.  Block id : %I64u\n", (uint64_t)imageInfo.ullBlockId);
#else
	printf("Received ImageAcquired event.  Block id : %lld\n", (long long unsigned int)imageInfo.ullBlockId);
#endif

	// Display the first 8 pixel values.
	printf("  ImageBuffer :");
	for (int i = 0; i < 8; i++)
	{
		printf(" %02X ", s_pucImgBuf[i]);
	}
	printf("...\n\n");

	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t DispFeature()
{
	uint32_t uiRaw;
	float64_t dExp, dGain, dFRate;


	// Exposure
	{
		// Get ExposureTimeRaw
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_EXPOSURE_TIME_RAW_I, 1, &uiRaw);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 200;
		}

		// Display ExposureTime. (In ExposureTime FeatureCSR, actual value can be calculated using normal multiply and divide calculation.)
		dExp = s_dExpBase * (double)uiRaw;
		printf(" Exposure time : %.4lf[ms]\n", dExp * s_dMultDspExp);
	}

	// Gain
	{
		// Get GainRaw
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_LMN_GAIN_RAW_I, 1, &uiRaw);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 201;
		}

		// Display Gain. (In Gain FeatureCSR, actual value can be calculated using normal multiply and divide calculation.)
		dGain = s_dGainBase * (double)uiRaw;
		printf(" Gain          : %.4f[db]\n", dGain);
	}

	// Framerate
	{
		// Get AcquisitionFrameRateRaw.
		s_uiStatus = Cam_ReadReg(s_hCam, RegMapBU::ADR_ACQ_FRAME_RATE_RAW_I, 1, &uiRaw);
		if (s_uiStatus != CAM_API_STS_SUCCESS)
		{
			return 202;
		}

		// Display Framerate. (In  AcquisitionFrameRate FeatureCSR, actual value can be calculated using normal multiply and divide calculation.)
		dFRate = s_dFrateBase * (double)uiRaw;
		printf(" Frame rate    : %.4f[fps]\n", dFRate);
	}

	return 0;
}

/////////////////////////////////////////////////////////
//
uint32_t ProcessLoop()
{
	int       iKey = 0;
	uint32_t  uiStatus = 0;
	bool8_t   bDisp = true;


	printf("\n");
	printf("ProcessLoop() started!\n");

	//
	while (1)
	{
		if (bDisp)
		{
			printf("\n");
			DispFeature();

			printf("---------------------------------------------------------------------\n");
			printf("Press '0' key to issue \"Software Trigger\" and grab a frame.\n");
			printf("Press '1' key to set exposure time\n");
			printf("Press '2' key to set gain\n");
			printf("Press '9' key to quit application.\n");
			printf("---------------------------------------------------------------------\n");

			bDisp = false;
		}

#if defined (_WIN32)
		iKey = _getch();
#else
		iKey = getchar();
#endif
		if (iKey == '0')
		{
			//
			uiStatus = ImageHandling();
			if (uiStatus != CAM_API_STS_SUCCESS)
			{
				break;
			}
			bDisp = true;
		} else if (iKey == '1')
		{
			//
			uiStatus = SetExposureTime();
			if (uiStatus != CAM_API_STS_SUCCESS)
			{
				break;
			}
			bDisp = true;
		} else if (iKey == '2')
		{
			//
			uiStatus = SetGain();
			if (uiStatus != CAM_API_STS_SUCCESS)
			{
				break;
			}
			bDisp = true;
		} else if (iKey == '9')
		{
			break;
		}
	}

	printf("ProcessLoop() finished!\n");
	return uiStatus;
}

